/*
 * Override Jupyter's default css inside elements with the hrepr class.
 */

.rendered_html .hrepr table {
	margin-top: 0px;
	margin-bottom: 0px;
	table-layout: auto;
}

.rendered_html tr, .rendered_html th, .rendered_html td {
	padding: 2px;
	text-align: left;
}
